@@ -5,11 +5,7 @@ from io import StringIO
 
 from . rules import tokenize
 
-
-amp_pattern = re.compile(r'&(amp;)?')
-lt_pattern = re.compile(r'<|(&lt;)')
-gt_pattern = re.compile(r'>|(&gt;)')
-
+token_regex = re.compile(r'<S/>|</?s>|<([wc]([\s]*|[\s]+[^>]*))>([^<]+)</[wc]>')
 
 def create_tei_tree():
     root = ET.Element('TEI')
@@ -22,23 +18,9 @@ def create_tei_tree():
 
 
 def index_of(string, substring, from_idx, val):
-    val[0] = ''
-    pattern = None
-    if substring == '&amp;':
-        pattern = amp_pattern
-    elif substring == '&lt;':
-        pattern = lt_pattern
-    elif substring == '&gt;':
-        pattern = gt_pattern
-   
-    if pattern:
-        for match in pattern.finditer(string):
-            if match.start() < from_idx:
-                continue
-            elif match.start() >= from_idx:
-                val[0] = match.group()
-                return match.start()
-        return -1
+    substring = re.sub(r'&amp;', '&', substring) 
+    substring = re.sub(r'&lt;', '<', substring) 
+    substring = re.sub(r'&gt;', '>', substring) 
     val[0] = substring
     return string.find(substring, from_idx)
 
@@ -53,15 +35,14 @@ def preprocess_tokens(tokens):
     org_text = []
     for match in stoken_regex.finditer(tokens):
         val = match.group()
-        val = val.replace('<s>', '')
-        val = val.replace('</s>', '')
-        val = val.replace('<w>', '')
-        val = val.replace('</w>', '')
-        val = val.replace('<c>', '')
-        val = val.replace('</c>', '')
-        val = val.replace('<S/>', ' ')
-        val = val.replace(' {2,}', ' ')
-        val = re.sub(' +', ' ', val)
+        val = re.sub(r'<s>', '', val)
+        val = re.sub(r'</s>', '', val)
+        val = re.sub(r'<[cw]([\s]*|[\s]+[^>]*)>', '', val)
+        val = re.sub(r'</w>', '', val)
+        val = re.sub(r'</c>', '', val)
+        val = re.sub(r'<S/>', ' ', val)
+        val = re.sub(r' {2,}', ' ', val)
+        val = re.sub(r' +', ' ', val)
         org_text.append(val.strip())
 
     return org_text
@@ -73,7 +54,6 @@ def process_tokenize_only(para, np, os):
     tokens = tokenize(para)
     org_text = preprocess_tokens(tokens)
 
-    token_regex = re.compile(r'<S/>|</?s>|<([wc])>([^<]+)</[wc]>')
     idx = 0
     ns = 1
     nt = 0
@@ -91,7 +71,7 @@ def process_tokenize_only(para, np, os):
         elif val == '<S/>':
             pass
         else:
-            val = match.group(2)
+            val = match.group(3)
             actual_val = ['']
             idx_of_token = index_of(para, val, idx, actual_val)
             if idx_of_token == -1:
@@ -108,8 +88,7 @@ def process_tokenize_only(para, np, os):
         os.write('\n')
 
 
-
-def process_conllu(para, np, os):
+def process_conllu(para, np, os, object_output=False):
     if para.startswith(u"\uFEFF"):
         para = para[1:]
 
@@ -117,11 +96,15 @@ def process_conllu(para, np, os):
 
     tokens = tokenize(para)
 
-    os.write('# newpar id = {}\n'.format(np))
+    if object_output:
+        metadata = '# newpar id = {}\n'.format(np)
+        doc = []
+        doc_sent = []
+    else:
+        os.write('# newpar id = {}\n'.format(np))
 
     org_text = preprocess_tokens(tokens)
 
-    token_regex = re.compile(r'<S/>|</?s>|<([wc])>([^<]+)</[wc]>')
     idx = 0
     ns = 1
     nt = 0
@@ -132,16 +115,26 @@ def process_conllu(para, np, os):
         if val == '<s>':
             nt = 0
             if ns != old_ns:
-                os.write('\n')
+                if not object_output:
+                    os.write('\n')
                 old_ns = ns
-            os.write('# sent_id = {}.{}\n'.format(np, ns))
-            os.write('# text = {}\n'.format(unescape_xml_chars(org_text[ns - 1])))
+            if object_output:
+                metadata += '# sent_id = {}.{}\n'.format(np, ns)
+                metadata += '# text = {}\n'.format(unescape_xml_chars(org_text[ns - 1]))
+            else:
+                os.write('# sent_id = {}.{}\n'.format(np, ns))
+                os.write('# text = {}\n'.format(unescape_xml_chars(org_text[ns - 1])))
         elif val == '</s>':
+            if object_output:
+                doc.append({'sentence': doc_sent, 'metadata': metadata})
+                doc_sent = []
+            metadata = ''
             ns += 1
         elif val == '<S/>':
             pass
         else:
-            val = match.group(2)
+            attribs = parse_attribs(match.group(2))
+            val = match.group(3)
             actual_val = ['']
             idx_of_token = index_of(para, val, idx, actual_val)
             if idx_of_token == -1:
@@ -149,18 +142,31 @@ def process_conllu(para, np, os):
             idx = max(idx, idx_of_token + len(actual_val[0]))
             idx_of_token += 1
             nt += 1
-            line = str(nt) + '\t{}\t_\t_\t_\t_\t_\t_\t_'.format(actual_val[0])
+
+            lemma = unescape_xml_chars(attribs.get('lemma', '_'))
+            xpos = attribs.get('xpos', '_')
+            upos = attribs.get('upos', '_')
             if idx < len(para) and not para_concat[idx].isspace():
                 space_after = 'SpaceAfter=No'
             else:
                 space_after = '_'
-            line += '\t{}\n'.format(space_after)
-            os.write(line)
+
+            if object_output:
+                tok = {'id': tuple([nt]), 'text': actual_val[0], 'lemma': lemma, 'xpos': xpos, 'upos': upos, 'misc': space_after}
+                doc_sent.append(tok)
+            else:
+                line = str(nt) + '\t{}\t{}\t{}\t{}\t_\t_\t_\t_\t{}\n'.format(actual_val[0], lemma,
+                                                            upos, xpos, space_after)
+                os.write(line)
             has_output = True
 
-    if has_output:
+    if has_output and not object_output:
         os.write('\n')
 
+    if object_output:
+        return doc
+    return
+
 
 def process_tei(para, np, os, tei_root):
     if para.startswith(u"\uFEFF"):
@@ -180,12 +186,9 @@ def process_tei(para, np, os, tei_root):
 
     org_text = preprocess_tokens(tokens)
 
-    token_regex = re.compile(r'<S/>|</?s>|<([wc])>([^<]+)</[wc]>')
     idx = 0
     ns = 1
     nt = 0
-    old_ns = 1
-    has_output = False
     for match in token_regex.finditer(tokens):
         val = match.group()
         if val == '<s>':
@@ -205,7 +208,8 @@ def process_tei(para, np, os, tei_root):
             parent_node.append(node)
             parent_map[node] = parent_node
         else:
-            val = match.group(2)
+            attribs = parse_attribs(match.group(2))
+            val = match.group(3)
             actual_val = ['']
             idx_of_token = index_of(para, val, idx, actual_val)
             if (idx_of_token == -1):
@@ -213,7 +217,7 @@ def process_tei(para, np, os, tei_root):
             idx = max(idx, idx_of_token + len(actual_val[0]))
             idx_of_token += 1
             nt += 1
-            if match.group(1) == 'c':
+            if match.group(1).startswith('c'):
                 tag_name = 'pc'
             else:
                 tag_name = 'w'
@@ -221,12 +225,35 @@ def process_tei(para, np, os, tei_root):
             node.text = actual_val[0]
             #node.set('xml:id', id_prefix + str(np) + '.' + str(ns) + '.t' + str(nt))
             node.attrib['{http://www.w3.org/XML/1998/namespace}id'] = id_prefix + str(np) + '.' + str(ns) + '.t' + str(nt)
+
+            lemma = attribs.get('lemma', None)
+            xpos = attribs.get('xpos', None)
+            upos = attribs.get('upos', None)
+            if lemma:
+                node.attrib['lemma'] = lemma
+            if xpos:
+                node.attrib['ana'] = 'mte:' + xpos
+            if upos:
+                node.attrib['msd'] = 'UposTag=' + upos
+
             parent_node.append(node)
             parent_map[node] = parent_node
 
 
-def process_text(text, os, tei_root, conllu, pass_newdoc_id):
+def parse_attribs(val):
+    res = dict()
+    for token in val.split():
+        key, val = token.split('=', 1)
+        if val[0] != '"' or val[-1] != '"':
+            raise Exception('Attrib value needs to be in double quotes.')
+        res[key] = val[1:-1]
+    return res
+
+
+def process_text(text, os, tei_root, conllu, pass_newdoc_id, object_output=False):
     np = 0
+    if object_output:
+        document = []
     for line in text:
         if line.isspace() or line == '':
             continue
@@ -236,12 +263,21 @@ def process_text(text, os, tei_root, conllu, pass_newdoc_id):
         line = normalize(line)
 
         if conllu:
+            if object_output:
+                pre_metadata = ''
             if pass_newdoc_id and line.startswith('# newdoc id = '):
                 np = 0
-                os.write(line)
-                os.write('\n')
+                if object_output:
+                    pre_metadata += line + '\n'
+                else:
+                    os.write(line)
+                    os.write('\n')
             np += 1
-            process_conllu(line, np, os)
+            out = process_conllu(line, np, os, object_output=object_output)
+            if object_output:
+                if pre_metadata and len(out) > 0:
+                    out[0]['metadata'] = pre_metadata + out[0]['metadata']
+                document.append(out)
         elif tei_root is not None:
             np += 1
             process_tei(line, np, os, tei_root)
@@ -249,12 +285,16 @@ def process_text(text, os, tei_root, conllu, pass_newdoc_id):
             np += 1
             process_tokenize_only(line, np, os)
 
+    if object_output:
+        return document
+    return None
+
 def normalize(text):
     text = text.replace('\xad', '-')   # Soft hyphens
     return text
 
 
-def run(text=None, in_file=None, in_files=None, out_file=None, to_stdout=False, tei=False, conllu=False, pass_newdoc_id=False):
+def run(text=None, in_file=None, in_files=None, out_file=None, to_stdout=False, tei=False, conllu=False, pass_newdoc_id=False, object_output=False):
     """
     Run Obeliks on specified input.
 
@@ -275,7 +315,11 @@ def run(text=None, in_file=None, in_files=None, out_file=None, to_stdout=False,
                         selected.
     """
     os = StringIO()
-    if to_stdout:
+    if object_output:
+        os = {}
+        conllu = True
+        tei = False
+    elif to_stdout:
         os = sys.stdout
     elif out_file is not None:
         os = open(out_file, 'w', encoding='utf-8')
@@ -295,21 +339,21 @@ def run(text=None, in_file=None, in_files=None, out_file=None, to_stdout=False,
 
     if tei:
         tei_tree = create_tei_tree()
-        process_text(text, os, tei_tree.getroot(), False, False)
+        out = process_text(text, os, tei_tree.getroot(), False, False)
         if isinstance(os, StringIO):
             return ET.tostring(tei_tree, encoding='utf8', method='xml')
         else:
             tei_tree.write(os.buffer, encoding='utf-8', xml_declaration=True, pretty_print=True)
     else:
-        process_text(text, os, None, conllu, pass_newdoc_id)
+        out = process_text(text, os, None, conllu, pass_newdoc_id, object_output=object_output)
     
     if isinstance(os, StringIO):
         contents = os.getvalue()
         os.close()
         return contents
 
-    if os is not sys.stdout:
+    if not object_output and os is not sys.stdout:
         os.close()
 
-    return None
+    return out
 
